/*
  JUICEROO Off-Grid solar EV charging controller 

  Written by Dominic Buchstaller.
  BSD license, check license.txt for more information
*/

#define SERIAL_PORT_HARDWARE Serial1
#define RS485_WB_TX_PIN 14
#define RS485_WB_RX_PIN 12
#define LCD_BL 22
#define RS485_INV_TX_PIN 26
#define RS485_INV_RX_PIN 27
#define LCD_BL_button_pin   19
#define request_switch_pin   16


#define modbus_delay 50
#define display_time 120000
#define wb_timout 5*60*1000 //5 minutes AC timout
#include <TFT_eSPI.h> // Graphics and font library for ST7735 driver chip
#include <SPI.h>
#include <ModbusMaster.h>               //Library for using ModbusMaster
#include "freefont.h"
#include "image_data.h"


TFT_eSPI display = TFT_eSPI();  // Invoke library, pins defined in User_Setup.h
#define TFT_W 128
#define TFT_H 128

// System states: Solar charging stationary battery, EV charging, paused EV charging
enum system_states {SOLAR, CHARGE, PAUSE};
system_states system_state;

bool request_switch, request_switch_last;

unsigned long last_button_press;
unsigned long last_wb_use;


uint16_t err_cnt_wb, err_cnt_inv;


// Using both hardware serial ports and two Modbus interface boards - may also work with one but why not
HardwareSerial mySerial1(1);
HardwareSerial mySerial2(2);

ModbusMaster node_wb,node_inv;                    //object node_inv for class ModbusMaster

void preTransmission()            //Function for setting stste of Pins DE & RE of RS-485
{
            
  
}

void postTransmission()
{
  
}


void setup()
{
  pinMode(LCD_BL_button_pin, INPUT_PULLDOWN);
  pinMode(request_switch_pin, INPUT_PULLDOWN);
  request_switch_last=digitalRead(request_switch_pin);
  pinMode(LCD_BL, OUTPUT);

  digitalWrite(LCD_BL,0);
  display.init();
  display.setSwapBytes(true);
  display.setRotation(3);
  
  display.fillScreen(0xFFFF);
  mySerial1.begin(9600, SERIAL_8N1,RS485_WB_RX_PIN,RS485_WB_TX_PIN);
  mySerial2.begin(9600, SERIAL_8N1,RS485_INV_RX_PIN,RS485_INV_TX_PIN);
  Serial.begin(9600);             //Baud Rate as 115200
  node_wb.begin(1, mySerial1);            //Slave ID as 1
  node_inv.begin(3, mySerial2);            //Slave ID as 2
  uint16_t err_cnt_inv=0;
  uint16_t err_cnt_wb=0;
  write_register_wb(0x4002,10);

  system_state=SOLAR;
  last_wb_use=millis();
  delay(100);

  

}

  
  
char incomingByte;

// Read wallbox register
uint16_t read_register_wb(uint16_t my_register)
{
   uint8_t result= node_wb.readHoldingRegisters(my_register,1);        //Writes value to 0x40000 holding register
   uint16_t my_result=0;
   if (result == node_wb.ku8MBSuccess)
   {
      my_result = node_wb.getResponseBuffer(0);
      //if ((my_result != 255) && (my_result != 0))
        Serial.println("Reading WB - Reg: " + String(my_register,HEX)+", result: " + String(my_result));
        err_cnt_wb=0;
   }
   else
   {
  
      err_cnt_wb++;
      Serial.println("WB Error: "+ String(result,HEX));
   }
   return(my_result);
}

// Write wallbox register
void write_register_wb(uint16_t my_register, uint16_t value)
{
   uint8_t result= node_wb.writeSingleRegister(my_register,value);        //Writes value to 0x40000 holding register
   delay(modbus_delay);
   
}

// Write inverter register
void write_register_inv(uint16_t my_register, uint16_t value)
{
   uint8_t result= node_inv.writeSingleRegister(my_register,value);        //Writes value to 0x40000 holding register
   delay(modbus_delay);
   
}

// Read multiple inverter register
uint16_t read_registers_inv(uint16_t base, uint16_t number)
{
   uint8_t result= node_inv.readInputRegisters(base,number);        //Writes value to 0x40000 holding register
   uint16_t my_result =0;
   if (result == node_inv.ku8MBSuccess)
   {
     my_result = 1;
     err_cnt_inv=0;
   }
   else
   {  my_result = 0;
      err_cnt_inv++;
      Serial.println("INV Error: "+ String(result,HEX));
   }
   delay(modbus_delay);
   return(my_result);
}

// Return single value from read response 
uint16_t read_register_inv(uint16_t my_register)
{
    uint16_t my_result = node_inv.getResponseBuffer(my_register);
    return(my_result);
}

uint16_t my_pv_power;
uint16_t my_wb_power;
uint16_t my_bat_power_l;
uint16_t my_bat_power_h;
uint16_t my_wb_power_l;
uint16_t my_wb_power_h;

int32_t my_bat_power;


uint16_t my_SOC;
uint16_t my_charger_current;
uint16_t wb_charger_current;


void update_display (float pow_car, float pow_sun, float pow_bat, uint16_t SOC, uint16_t wb_fault, uint16_t inv_fault)
{
  #define y1offset 55
  #define y2offset 85
  #define xoffset 0
  
 // If we lose communication to wallbox or inverter we display numeric values in RED
 // If component is idle of off we gray them out

  display.setSwapBytes(true);
  if (system_state==CHARGE)
    display.pushImage(0,0,60,60,car);
  else
    display.pushImage(0,0,60,60,car_off);

  if (pow_sun==0)  
    display.pushImage(68,0,60,60,sun_off);
  else
    display.pushImage(68,0,60,60,sun);

  if (my_SOC<30)
    display.pushImage(68,75,60,60,battery_off);
  else
    display.pushImage(68,75,60,60,battery);
  display.setSwapBytes(false);

   display.fillRect(xoffset, y1offset, 128, 20, 0xFFFF);
   
   display.fillRect(xoffset+0, y2offset, 60, 50, 0xFFFF);
  display.setFreeFont(FSSB9);
  
  
  if (wb_fault)
    display.setTextColor(TFT_RED, TFT_WHITE);
  else
    display.setTextColor(TFT_BLACK, TFT_WHITE);

  display.drawString(String(pow_car/1000.0,2)+String("kW"), xoffset, y1offset+2, GFXFF);

  if (inv_fault)
    display.setTextColor(TFT_RED, TFT_WHITE);
  else
    display.setTextColor(TFT_BLACK, TFT_WHITE);
  display.drawString(String(pow_sun/1000.0,2)+String("kW"), xoffset+65, y1offset+2, GFXFF);
  display.drawString(String(pow_bat/1000.0,2)+String("kW"), xoffset, y2offset+2, GFXFF);
  display.drawString(SOC+String("%"), xoffset+2, y2offset+20, GFXFF);
}

void loop()
{
  
  // Read data from WB if  powered on
  //if (system_state==CHARGE)
  //  wb_charger_current=read_register_wb(0x4002);
 
  //  Serial.println("SW Version RFID/ModBUS" + String(read_register_wb(0x30100)));
  //  delay(modbus_delay);
  //  Serial.println("SW Version Mainboard" + String(read_register_wb(0x30101)));
  //  delay(modbus_delay);
   
  // Read data from inverter
  if (read_registers_inv(0,20))
  {
    my_SOC = read_register_inv(18);
    my_pv_power = read_register_inv(04)/10;

  }
  uint8_t this_base=69;
  if (read_registers_inv(this_base,10))
  {
    my_bat_power_h = read_register_inv(77-this_base);
    my_bat_power_l = read_register_inv(78-this_base);
    my_bat_power = -((my_bat_power_h<<16) | my_bat_power_l)/10;
    my_wb_power_h = read_register_inv(69-this_base);
    my_wb_power_l = read_register_inv(70-this_base);
    my_wb_power = ((my_wb_power_h<<16) | my_wb_power_l)/10;
    if (my_wb_power>4000)
      my_wb_power=4000;
    delay(modbus_delay);
  }

  update_display(my_wb_power,my_pv_power,my_bat_power,my_SOC,err_cnt_wb,err_cnt_inv);

  // Check if charging request switch was flipped (does not matter if on or off) 
  request_switch=digitalRead(request_switch_pin);
  if (request_switch!=request_switch_last)
  {
    request_switch_last=request_switch;
    system_state=CHARGE;
    last_wb_use=millis();
  }

  // Check if wallbox is being used and update timout
  if ((my_wb_power>50))
    last_wb_use=millis();

  // Check if wallbox is idle for more than wb_timout -> chargin request has been fulfilled
  if ((millis()-last_wb_use)>wb_timout)
  {
    if (system_state==CHARGE)
      system_state=SOLAR;
  }

  // Resume charging automatically if SOC is sufficient
  if (system_state==PAUSE)
  {
    // Battery full - start/resume charging
    if (my_SOC>90)
    {
      system_state=CHARGE;
      last_wb_use=millis();
    }

    // Battery almost empty but sufficient PV power - start/resume charging
    if ((my_SOC>40) && (my_pv_power>2500))
    {
      system_state=CHARGE;
      last_wb_use=millis();
    }
      
  }

  // Batter empty -> pause charging reagardless of requests
  if ((system_state==CHARGE) && (my_SOC<31))
      system_state=PAUSE;

  // We only write commands to the inverter on changes - will beep otherwise
  // if (system_state!=system_state_last)
  // {
  //   system_state_last=system_state;
  //   if (system_state==CHARGE)
  //   {
  //     write_register_inv(0x0, 0x0000 );
  //     Serial.println("Send ON command to inverter");
  //   }
  //   else
  //   {
  //     write_register_inv(0x0, 0x0100 );
  //     Serial.println("Send OFF command to inverter");
  //   }
      
  // }
  // We write commands to the inverter
  if ((system_state==CHARGE) && (my_wb_power<2))
    write_register_inv(0x0, 0x0000 );
  if ((system_state!=CHARGE) && (my_wb_power>0))
    write_register_inv(0x0, 0x0100 );

  // Available charging current calculated from current PV power
  my_charger_current=my_pv_power/230;

  // Limit current 6-16A if SOC < 80
  
  if (my_charger_current > 16)
    my_charger_current=16;
  if (my_charger_current < 6)
    my_charger_current=6;

  // If battery is full - charge full power
  if (my_SOC>80)
    my_charger_current=16;
    
  // Write charging current limit to walbox only if charging (wallbox does not like Modbus commands during charging handshake)
  if (my_wb_power>1000)
    write_register_wb(0x4002,my_charger_current);
    
  // If wallbox is powered but we cannot communicate -> turn it off and on again
  if ((system_state==CHARGE) && (err_cnt_wb>20))
  {
    err_cnt_wb=0;
    Serial.println("Restarting WB");
    write_register_inv(0x0, 0x0100);
    delay(3000);
    write_register_inv(0x0, 0x0000);
  }

  //Serial.println("Request switch: " + String(request_switch));
  Serial.println("System State (last) {SOLAR, CHARGE, PAUSE}: " + String(system_state));
  
  Serial.println("SOC:" + String(my_SOC)+ ", PV [W]:" + String(my_pv_power)+ ", WB [W]:" + String(my_wb_power)+ ", BAT [W]:" + String(my_bat_power)+ ", EV set[A]:" + String(my_charger_current));
  Serial.println("EV get[A]:" + String(wb_charger_current));
  Serial.println("err_cnt_inv:" + String(err_cnt_inv)+ ", err_cnt_wb:" + String(err_cnt_wb));
  Serial.println("Display Timeout: " + String((millis()-last_button_press)/1000));
  Serial.println("AC Timeout: " + String((millis()-last_wb_use)/1000));  

  unsigned long start_time=millis();
  while ((millis()-start_time)<3000)
  {
    delay(100);
    // Light up LCD display if display button is pressed
    if (digitalRead(LCD_BL_button_pin))
    {
      digitalWrite(LCD_BL,1);
      last_button_press=millis();
    }
    if ((millis()-last_button_press)>display_time)
      digitalWrite(LCD_BL,0);
  }

}